@@ -12,7 +12,7 @@ NORETURN(static void raise_argument_error(rb_execution_context_t *ec, const rb_i
 NORETURN(static void argument_arity_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const int miss_argc, const int min_argc, const int max_argc));
 NORETURN(static void argument_kw_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const char *error, const VALUE keys));
 VALUE rb_keyword_error_new(const char *error, VALUE keys); /* class.c */
-static VALUE method_missing(VALUE obj, ID id, int argc, const VALUE *argv,
+static VALUE method_missing(rb_execution_context_t *ec, VALUE obj, ID id, int argc, const VALUE *argv,
                             enum method_missing_reason call_status, int kw_splat);
 #if !defined(_MSC_VER) || !defined(MJIT_HEADER)
 MJIT_FUNC_EXPORTED
@@ -27,7 +27,7 @@ struct args_info {
     /* additional args info */
     int rest_index;
     int rest_dupped;
-    const struct rb_call_info_kw_arg *kw_arg;
+    const struct rb_callinfo_kwarg *kw_arg;
     VALUE *kw_argv;
     VALUE rest;
 };
@@ -187,128 +187,10 @@ args_rest_array(struct args_info *args)
     return ary;
 }
 
-#define KW_HASH_HAS_NO_KEYS 0
-#define KW_HASH_HAS_SYMBOL_KEY 1
-#define KW_HASH_HAS_OTHER_KEY 2
-#define KW_HASH_HAS_BOTH_KEYS 3
-
-static int
-keyword_hash_symbol_other_iter(st_data_t key, st_data_t val, st_data_t arg)
-{
-    *(int*)arg |= SYMBOL_P((VALUE)key) ? KW_HASH_HAS_SYMBOL_KEY : KW_HASH_HAS_OTHER_KEY;
-
-    if ((*(int*)arg & KW_HASH_HAS_BOTH_KEYS) == KW_HASH_HAS_BOTH_KEYS) {
-	return ST_STOP;
-    }
-
-    return ST_CONTINUE;
-}
-
-static int
-keyword_hash_symbol_other(VALUE hash)
-{
-    int symbol_other = KW_HASH_HAS_NO_KEYS;
-    rb_hash_stlike_foreach(hash, keyword_hash_symbol_other_iter, (st_data_t)(&symbol_other));
-    return symbol_other;
-}
-
-static int
-keyword_hash_split_iter(st_data_t key, st_data_t val, st_data_t arg)
-{
-    if (SYMBOL_P((VALUE)key)) {
-	rb_hash_aset((VALUE)arg, (VALUE)key, (VALUE)val);
-	return ST_DELETE;
-    }
-
-    return ST_CONTINUE;
-}
-
-static void
-keyword_hash_split(VALUE *kw_hash_ptr, VALUE *rest_hash_ptr)
-{
-    *kw_hash_ptr = rb_hash_new();
-    rb_hash_stlike_foreach(*rest_hash_ptr, keyword_hash_split_iter, (st_data_t)(*kw_hash_ptr));
-}
-
-static int
-keyword_hash_p(VALUE *kw_hash_ptr, VALUE *rest_hash_ptr, int check_only_symbol)
-{
-    *rest_hash_ptr = rb_check_hash_type(*kw_hash_ptr);
-
-    if (!NIL_P(*rest_hash_ptr)) {
-	if (check_only_symbol) {
-	    switch (keyword_hash_symbol_other(*rest_hash_ptr)) {
-              case KW_HASH_HAS_NO_KEYS:
-              case KW_HASH_HAS_SYMBOL_KEY:
-		break;
-              case KW_HASH_HAS_OTHER_KEY:
-		*kw_hash_ptr = Qnil;
-		return FALSE;
-              case KW_HASH_HAS_BOTH_KEYS:
-                *rest_hash_ptr = rb_hash_dup(*rest_hash_ptr);
-		keyword_hash_split(kw_hash_ptr, rest_hash_ptr);
-		return TRUE;
-	    }
-	}
-	*kw_hash_ptr = *rest_hash_ptr;
-	*rest_hash_ptr = Qfalse;
-	return TRUE;
-    }
-    else {
-	*kw_hash_ptr = Qnil;
-	return FALSE;
-    }
-}
-
-static VALUE
-args_pop_keyword_hash(struct args_info *args, VALUE *kw_hash_ptr, int check_only_symbol)
-{
-    VALUE rest_hash;
-
-    if (args->rest == Qfalse) {
-      from_argv:
-	VM_ASSERT(args->argc > 0);
-	*kw_hash_ptr = args->argv[args->argc-1];
-
-	if (keyword_hash_p(kw_hash_ptr, &rest_hash, check_only_symbol)) {
-	    if (rest_hash) {
-		args->argv[args->argc-1] = rest_hash;
-	    }
-	    else {
-		args->argc--;
-		return TRUE;
-	    }
-	}
-    }
-    else {
-	long len = RARRAY_LEN(args->rest);
-
-	if (len > 0) {
-	    *kw_hash_ptr = RARRAY_AREF(args->rest, len - 1);
-
-	    if (keyword_hash_p(kw_hash_ptr, &rest_hash, check_only_symbol)) {
-		if (rest_hash) {
-		    RARRAY_ASET(args->rest, len - 1, rest_hash);
-		}
-		else {
-		    arg_rest_dup(args);
-		    rb_ary_pop(args->rest);
-		    return TRUE;
-		}
-	    }
-	}
-	else {
-	    goto from_argv;
-	}
-    }
-
-    return FALSE;
-}
-
 static int
 args_kw_argv_to_hash(struct args_info *args)
 {
-    const struct rb_call_info_kw_arg *kw_arg = args->kw_arg;
+    const struct rb_callinfo_kwarg *kw_arg = args->kw_arg;
     const VALUE *const passed_keywords = kw_arg->keywords;
     const int kw_len = kw_arg->keyword_len;
     VALUE h = rb_hash_new_with_size(kw_len);
@@ -326,29 +208,6 @@ args_kw_argv_to_hash(struct args_info *args)
     return args->argc;
 }
 
-static void
-args_stored_kw_argv_to_hash(struct args_info *args)
-{
-    int i;
-    const struct rb_call_info_kw_arg *kw_arg = args->kw_arg;
-    const VALUE *const passed_keywords = kw_arg->keywords;
-    const int passed_keyword_len = kw_arg->keyword_len;
-    VALUE h = rb_hash_new_with_size(passed_keyword_len);
-
-    for (i=0; i<passed_keyword_len; i++) {
-	rb_hash_aset(h, passed_keywords[i], args->kw_argv[i]);
-    }
-    args->kw_argv = NULL;
-
-    if (args->rest) {
-	arg_rest_dup(args);
-	rb_ary_push(args->rest, h);
-    }
-    else {
-	args->argv[args->argc++] = h;
-    }
-}
-
 static inline void
 args_setup_lead_parameters(struct args_info *args, int argc, VALUE *locals)
 {
@@ -541,9 +400,15 @@ args_setup_kw_parameters(rb_execution_context_t *const ec, const rb_iseq_t *cons
 }
 
 static inline void
-args_setup_kw_rest_parameter(VALUE keyword_hash, VALUE *locals)
+args_setup_kw_rest_parameter(VALUE keyword_hash, VALUE *locals, int kw_flag)
 {
-    locals[0] = NIL_P(keyword_hash) ? rb_hash_new() : rb_hash_dup(keyword_hash);
+    if (NIL_P(keyword_hash)) {
+        keyword_hash = rb_hash_new();
+    }
+    else if (!(kw_flag & VM_CALL_KW_SPLAT_MUT)) {
+        keyword_hash = rb_hash_dup(keyword_hash);
+    }
+    locals[0] = keyword_hash;
 }
 
 static inline void
@@ -570,171 +435,40 @@ fill_keys_values(st_data_t key, st_data_t val, st_data_t ptr)
 }
 
 static inline int
-ignore_keyword_hash_p(VALUE keyword_hash, const rb_iseq_t * const iseq)
+ignore_keyword_hash_p(VALUE keyword_hash, const rb_iseq_t * const iseq, unsigned int * kw_flag, VALUE * converted_keyword_hash)
 {
-    if (!(iseq->body->param.flags.has_kw) &&
-	      !(iseq->body->param.flags.has_kwrest)) {
-	keyword_hash = rb_check_hash_type(keyword_hash);
-
-	if (!NIL_P(keyword_hash) && RHASH_EMPTY_P(keyword_hash)) {
-	    return 1;
-	}
+    if (!RB_TYPE_P(keyword_hash, T_HASH)) {
+        keyword_hash = rb_to_hash_type(keyword_hash);
     }
-
-    return 0;
-}
-
-VALUE rb_iseq_location(const rb_iseq_t *iseq);
-
-/* -- Remove In 3.0 -- */
-
-/* This is a map from caller PC to a set of callee methods.
- * When a warning about keyword argument change is printed,
- * it keeps the pair of callee and caller.
- */
-static st_table *caller_to_callees = 0;
-
-static VALUE
-rb_warn_check(const rb_execution_context_t * const ec, const rb_iseq_t *const iseq)
-{
-    if (!rb_warning_category_enabled_p(RB_WARN_CATEGORY_DEPRECATED)) return 1;
-
-    if (!iseq) return 0;
-
-    const st_data_t callee = (st_data_t)(iseq->body->iseq_unique_id * 2);
-
-    const rb_control_frame_t * const cfp = rb_vm_get_ruby_level_next_cfp(ec, ec->cfp);
-
-    if (!cfp) return 0;
-
-    const st_data_t caller = (st_data_t)cfp->pc;
-
-    if (!caller_to_callees) {
-        caller_to_callees = st_init_numtable();
-    }
-
-    st_data_t val;
-    if (st_lookup(caller_to_callees, caller, &val)) {
-        st_table *callees;
-
-        if (val & 1) {
-            val &= ~(st_data_t)1;
-            if (val == callee) return 1; /* already warned */
-
-            callees = st_init_numtable();
-            st_insert(callees, val, 1);
-        }
-        else {
-            callees = (st_table *) val;
-            if (st_is_member(callees, callee)) return 1; /* already warned */
-        }
-        st_insert(callees, callee, 1);
-        st_insert(caller_to_callees, caller, (st_data_t) callees);
-    }
-    else {
-        st_insert(caller_to_callees, caller, callee | 1);
-    }
-
-    return 0; /* not warned yet for the pair of caller and callee */
-}
-
-static inline void
-rb_warn_keyword_to_last_hash(rb_execution_context_t * const ec, struct rb_calling_info *calling, const struct rb_call_info *ci, const rb_iseq_t * const iseq)
-{
-    if (rb_warn_check(ec, iseq)) return;
-
-    VALUE name, loc;
-    if (calling->recv == Qundef) {
-        rb_warn("Passing the keyword argument as the last hash parameter is deprecated");
-        return;
-    }
-    name = rb_id2str(ci->mid);
-    loc = rb_iseq_location(iseq);
-    if (NIL_P(loc)) {
-        rb_warn("Passing the keyword argument for `%"PRIsVALUE"' as the last hash parameter is deprecated",
-                name);
-    }
-    else {
-        rb_warn("Passing the keyword argument as the last hash parameter is deprecated");
-        if (name) {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method `%"PRIsVALUE"' is defined here", name);
-        }
-        else {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method is defined here");
-        }
-    }
-}
-
-static inline void
-rb_warn_split_last_hash_to_keyword(rb_execution_context_t * const ec, struct rb_calling_info *calling, const struct rb_call_info *ci, const rb_iseq_t * const iseq)
-{
-    if (rb_warn_check(ec, iseq)) return;
-
-    VALUE name, loc;
-    name = rb_id2str(ci->mid);
-    loc = rb_iseq_location(iseq);
-    if (NIL_P(loc)) {
-        rb_warn("Splitting the last argument for `%"PRIsVALUE"' into positional and keyword parameters is deprecated",
-                name);
-    }
-    else {
-        rb_warn("Splitting the last argument into positional and keyword parameters is deprecated");
-        if (calling->recv != Qundef) {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method `%"PRIsVALUE"' is defined here", name);
-        }
-        else {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method is defined here");
-        }
-    }
-}
-
-static inline void
-rb_warn_last_hash_to_keyword(rb_execution_context_t * const ec, struct rb_calling_info *calling, const struct rb_call_info *ci, const rb_iseq_t * const iseq)
-{
-    if (rb_warn_check(ec, iseq)) return;
-
-    VALUE name, loc;
-    name = rb_id2str(ci->mid);
-    loc = rb_iseq_location(iseq);
-    if (NIL_P(loc)) {
-        rb_warn("Using the last argument for `%"PRIsVALUE"' as keyword parameters is deprecated; maybe ** should be added to the call",
-                name);
-    }
-    else {
-        rb_warn("Using the last argument as keyword parameters is deprecated; maybe ** should be added to the call");
-        if (calling->recv != Qundef) {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method `%"PRIsVALUE"' is defined here", name);
-        }
-        else {
-            rb_compile_warn(RSTRING_PTR(RARRAY_AREF(loc, 0)), FIX2INT(RARRAY_AREF(loc, 1)),
-                            "The called method is defined here");
-        }
+    if (!(*kw_flag & VM_CALL_KW_SPLAT_MUT) &&
+            (iseq->body->param.flags.has_kwrest ||
+             iseq->body->param.flags.ruby2_keywords)) {
+        *kw_flag |= VM_CALL_KW_SPLAT_MUT;
+        keyword_hash = rb_hash_dup(keyword_hash);
     }
+    *converted_keyword_hash = keyword_hash;
+    return !(iseq->body->param.flags.has_kw) &&
+           !(iseq->body->param.flags.has_kwrest) &&
+           RHASH_EMPTY_P(keyword_hash);
 }
 
 static int
 setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * const iseq,
 			 struct rb_calling_info *const calling,
-			 const struct rb_call_info *ci,
-			 VALUE * const locals, const enum arg_setup_type arg_setup_type)
+                         const struct rb_callinfo *ci,
+                         VALUE * const locals, const enum arg_setup_type arg_setup_type)
 {
     const int min_argc = iseq->body->param.lead_num + iseq->body->param.post_num;
     const int max_argc = (iseq->body->param.flags.has_rest == FALSE) ? min_argc + iseq->body->param.opt_num : UNLIMITED_ARGUMENTS;
-    int opt_pc = 0;
     int given_argc;
-    int kw_splat = FALSE;
-    unsigned int kw_flag = ci->flag & (VM_CALL_KWARG | VM_CALL_KW_SPLAT);
+    unsigned int kw_flag = vm_ci_flag(ci) & (VM_CALL_KWARG | VM_CALL_KW_SPLAT | VM_CALL_KW_SPLAT_MUT);
+    int opt_pc = 0, allow_autosplat = !kw_flag;
     struct args_info args_body, *args;
     VALUE keyword_hash = Qnil;
     VALUE * const orig_sp = ec->cfp->sp;
     unsigned int i;
-    int remove_empty_keyword_hash = 1;
     VALUE flag_keyword_hash = 0;
+    VALUE converted_keyword_hash = 0;
 
     vm_check_canary(ec, orig_sp);
     /*
@@ -763,7 +497,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     args->rest_dupped = FALSE;
 
     if (kw_flag & VM_CALL_KWARG) {
-	args->kw_arg = ((struct rb_call_info_with_kwarg *)ci)->kw_arg;
+	args->kw_arg = vm_ci_kwarg(ci);
 
 	if (iseq->body->param.flags.has_kw) {
 	    int kw_len = args->kw_arg->keyword_len;
@@ -776,7 +510,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	else {
 	    args->kw_argv = NULL;
 	    given_argc = args_kw_argv_to_hash(args);
-	    kw_flag |= VM_CALL_KW_SPLAT;
+            kw_flag |= VM_CALL_KW_SPLAT | VM_CALL_KW_SPLAT_MUT;
 	}
     }
     else {
@@ -784,11 +518,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	args->kw_argv = NULL;
     }
 
-    if (kw_flag && iseq->body->param.flags.ruby2_keywords) {
-        remove_empty_keyword_hash = 0;
-    }
-
-    if (ci->flag & VM_CALL_ARGS_SPLAT) {
+    if (vm_ci_flag(ci) & VM_CALL_ARGS_SPLAT) {
         VALUE rest_last = 0;
         int len;
 	args->rest = locals[--args->argc];
@@ -801,10 +531,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
             if (RB_TYPE_P(rest_last, T_HASH) &&
                 (((struct RHash *)rest_last)->basic.flags & RHASH_PASS_AS_KEYWORDS)) {
                 rest_last = rb_hash_dup(rest_last);
-                kw_flag |= VM_CALL_KW_SPLAT;
-                if (iseq->body->param.flags.ruby2_keywords) {
-                    remove_empty_keyword_hash = 0;
-                }
+                kw_flag |= VM_CALL_KW_SPLAT | VM_CALL_KW_SPLAT_MUT;
             }
             else {
                 rest_last = 0;
@@ -812,47 +539,53 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
         }
 
         if (kw_flag & VM_CALL_KW_SPLAT) {
-            if (len > 0 && ignore_keyword_hash_p(rest_last, iseq)) {
-                if (given_argc != min_argc) {
-                    if (remove_empty_keyword_hash) {
-                        arg_rest_dup(args);
-                        rb_ary_pop(args->rest);
-                        given_argc--;
-                        kw_flag &= ~VM_CALL_KW_SPLAT;
-                    }
-                    else {
-                        flag_keyword_hash = rest_last;
-                    }
-                }
-                else {
-                    rb_warn_keyword_to_last_hash(ec, calling, ci, iseq);
-                }
+            if (ignore_keyword_hash_p(rest_last, iseq, &kw_flag, &converted_keyword_hash)) {
+                arg_rest_dup(args);
+                rb_ary_pop(args->rest);
+                given_argc--;
+                kw_flag &= ~(VM_CALL_KW_SPLAT | VM_CALL_KW_SPLAT_MUT);
 	    }
-            else if (!remove_empty_keyword_hash && rest_last) {
-                flag_keyword_hash = rest_last;
+            else {
+                if (rest_last != converted_keyword_hash) {
+                    rest_last = converted_keyword_hash;
+                    arg_rest_dup(args);
+                    RARRAY_ASET(args->rest, len - 1, rest_last);
+                }
+
+                if (iseq->body->param.flags.ruby2_keywords && rest_last) {
+                    flag_keyword_hash = rest_last;
+                }
+                else if (iseq->body->param.flags.has_kw || iseq->body->param.flags.has_kwrest) {
+                    arg_rest_dup(args);
+                    rb_ary_pop(args->rest);
+                    given_argc--;
+                    keyword_hash = rest_last;
+                }
             }
         }
     }
     else {
         if (kw_flag & VM_CALL_KW_SPLAT) {
             VALUE last_arg = args->argv[args->argc-1];
-            if (ignore_keyword_hash_p(last_arg, iseq)) {
-                if (given_argc != min_argc) {
-                    if (remove_empty_keyword_hash) {
-                        args->argc--;
-                        given_argc--;
-                        kw_flag &= ~VM_CALL_KW_SPLAT;
-                    }
-                    else {
-                        flag_keyword_hash = last_arg;
-                    }
-                }
-                else {
-                    rb_warn_keyword_to_last_hash(ec, calling, ci, iseq);
-                }
+            if (ignore_keyword_hash_p(last_arg, iseq, &kw_flag, &converted_keyword_hash)) {
+                args->argc--;
+                given_argc--;
+                kw_flag &= ~(VM_CALL_KW_SPLAT | VM_CALL_KW_SPLAT_MUT);
 	    }
-            else if (!remove_empty_keyword_hash) {
-                flag_keyword_hash = args->argv[args->argc-1];
+            else {
+                if (last_arg != converted_keyword_hash) {
+                    last_arg = converted_keyword_hash;
+                    args->argv[args->argc-1] = last_arg;
+                }
+
+                if (iseq->body->param.flags.ruby2_keywords) {
+                    flag_keyword_hash = last_arg;
+                }
+                else if (iseq->body->param.flags.has_kw || iseq->body->param.flags.has_kwrest) {
+                    args->argc--;
+                    given_argc--;
+                    keyword_hash = last_arg;
+                }
             }
         }
 	args->rest = Qfalse;
@@ -866,13 +599,14 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	rb_raise(rb_eArgError, "no keywords accepted");
     }
 
+
     switch (arg_setup_type) {
       case arg_setup_method:
 	break; /* do nothing special */
       case arg_setup_block:
-	if (given_argc == 1 &&
-	    (min_argc > 0 || iseq->body->param.opt_num > 1 ||
-	     iseq->body->param.flags.has_kw || iseq->body->param.flags.has_kwrest) &&
+        if (given_argc == (NIL_P(keyword_hash) ? 1 : 2) &&
+            allow_autosplat &&
+            (min_argc > 0 || iseq->body->param.opt_num > 1) &&
 	    !iseq->body->param.flags.ambiguous_param0 &&
 	    args_check_block_arg0(args)) {
 	    given_argc = RARRAY_LENINT(args->rest);
@@ -882,60 +616,13 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 
     /* argc check */
     if (given_argc < min_argc) {
-	if (given_argc == min_argc - 1 && args->kw_argv) {
-	    args_stored_kw_argv_to_hash(args);
-	    given_argc = args_argc(args);
-	}
-	else {
-	    if (arg_setup_type == arg_setup_block) {
-		CHECK_VM_STACK_OVERFLOW(ec->cfp, min_argc);
-		given_argc = min_argc;
-		args_extend(args, min_argc);
-	    }
-	    else {
-		argument_arity_error(ec, iseq, given_argc, min_argc, max_argc);
-	    }
-	}
-    }
-
-    if (kw_flag & VM_CALL_KW_SPLAT) {
-	kw_splat = !iseq->body->param.flags.has_rest;
-    }
-    if ((iseq->body->param.flags.has_kw || iseq->body->param.flags.has_kwrest ||
-	 (kw_splat && given_argc > max_argc)) &&
-	args->kw_argv == NULL) {
-        if (given_argc > min_argc) {
-            if (kw_flag) {
-                int check_only_symbol = (kw_flag & VM_CALL_KW_SPLAT) &&
-                                        iseq->body->param.flags.has_kw &&
-                                        !iseq->body->param.flags.has_kwrest;
-
-                if (args_pop_keyword_hash(args, &keyword_hash, check_only_symbol)) {
-                    given_argc--;
-                }
-                else if (check_only_symbol) {
-                    if (keyword_hash != Qnil) {
-                        rb_warn_split_last_hash_to_keyword(ec, calling, ci, iseq);
-                    }
-                    else {
-                        rb_warn_keyword_to_last_hash(ec, calling, ci, iseq);
-                    }
-                }
-            }
-            else if (args_pop_keyword_hash(args, &keyword_hash, 1)) {
-                /* Warn the following:
-                 * def foo(k:1) p [k]; end
-                 * foo({k:42}) #=> 42
-                 */
-                rb_warn_last_hash_to_keyword(ec, calling, ci, iseq);
-                given_argc--;
-            }
-            else if (keyword_hash != Qnil) {
-                rb_warn_split_last_hash_to_keyword(ec, calling, ci, iseq);
-            }
+        if (arg_setup_type == arg_setup_block) {
+            CHECK_VM_STACK_OVERFLOW(ec->cfp, min_argc);
+            given_argc = min_argc;
+            args_extend(args, min_argc);
         }
-        else if (given_argc == min_argc && kw_flag) {
-            rb_warn_keyword_to_last_hash(ec, calling, ci, iseq);
+        else {
+            argument_arity_error(ec, iseq, given_argc, min_argc, max_argc);
         }
     }
 
@@ -975,7 +662,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	VALUE * const klocals = locals + iseq->body->param.keyword->bits_start - iseq->body->param.keyword->num;
 
 	if (args->kw_argv != NULL) {
-	    const struct rb_call_info_kw_arg *kw_arg = args->kw_arg;
+	    const struct rb_callinfo_kwarg *kw_arg = args->kw_arg;
 	    args_setup_kw_parameters(ec, iseq, args->kw_argv, kw_arg->keyword_len, kw_arg->keywords, klocals);
 	}
 	else if (!NIL_P(keyword_hash)) {
@@ -995,7 +682,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
 	}
     }
     else if (iseq->body->param.flags.has_kwrest) {
-	args_setup_kw_rest_parameter(keyword_hash, locals + iseq->body->param.keyword->rest_start);
+        args_setup_kw_rest_parameter(keyword_hash, locals + iseq->body->param.keyword->rest_start, kw_flag);
     }
     else if (!NIL_P(keyword_hash) && RHASH_SIZE(keyword_hash) > 0 && arg_setup_type == arg_setup_method) {
 	argument_kw_error(ec, iseq, "unknown", rb_hash_keys(keyword_hash));
@@ -1014,7 +701,7 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     {
 	int i;
 	for (i=0; i<iseq->body->param.size; i++) {
-	    fprintf(stderr, "local[%d] = %p\n", i, (void *)locals[i]);
+	    ruby_debug_printf("local[%d] = %p\n", i, (void *)locals[i]);
 	}
     }
 #endif
@@ -1023,8 +710,6 @@ setup_parameters_complex(rb_execution_context_t * const ec, const rb_iseq_t * co
     return opt_pc;
 }
 
-void rb_backtrace_use_iseq_first_lineno_for_last_location(VALUE self); /* vm_backtrace.c */
-
 static void
 raise_argument_error(rb_execution_context_t *ec, const rb_iseq_t *iseq, const VALUE exc)
 {
@@ -1103,11 +788,10 @@ vm_caller_setup_arg_splat(rb_control_frame_t *cfp, struct rb_calling_info *calli
 }
 
 static inline void
-vm_caller_setup_arg_kw(rb_control_frame_t *cfp, struct rb_calling_info *calling, const struct rb_call_info *ci)
+vm_caller_setup_arg_kw(rb_control_frame_t *cfp, struct rb_calling_info *calling, const struct rb_callinfo *ci)
 {
-    struct rb_call_info_with_kwarg *ci_kw = (struct rb_call_info_with_kwarg *)ci;
-    const VALUE *const passed_keywords = ci_kw->kw_arg->keywords;
-    const int kw_len = ci_kw->kw_arg->keyword_len;
+    const VALUE *const passed_keywords = vm_ci_kwarg(ci)->keywords;
+    const int kw_len = vm_ci_kwarg(ci)->keyword_len;
     const VALUE h = rb_hash_new_with_size(kw_len);
     VALUE *sp = cfp->sp;
     int i;
@@ -1155,13 +839,11 @@ refine_sym_proc_call(RB_BLOCK_CALL_FUNC_ARGLIST(yielded_arg, callback_arg))
 {
     VALUE obj;
     ID mid;
-    const rb_callable_method_entry_t *me;
+    const rb_callable_method_entry_t *me = 0; /* for hidden object case */
     rb_execution_context_t *ec;
     const VALUE symbol = RARRAY_AREF(callback_arg, 0);
     const VALUE refinements = RARRAY_AREF(callback_arg, 1);
     int kw_splat = RB_PASS_CALLED_KEYWORDS;
-    VALUE v;
-    VALUE ret;
     VALUE klass;
 
     if (argc-- < 1) {
@@ -1182,22 +864,17 @@ refine_sym_proc_call(RB_BLOCK_CALL_FUNC_ARGLIST(yielded_arg, callback_arg))
     if (!NIL_P(blockarg)) {
 	vm_passed_block_handler_set(ec, blockarg);
     }
-    v = rb_adjust_argv_kw_splat(&argc, &argv, &kw_splat);
     if (!me) {
-        ret = method_missing(obj, mid, argc, argv, MISSING_NOENTRY, kw_splat);
+        return method_missing(ec, obj, mid, argc, argv, MISSING_NOENTRY, kw_splat);
     }
-    else {
-        ret = rb_vm_call0(ec, obj, mid, argc, argv, me, kw_splat);
-    }
-    rb_free_tmp_buffer(&v);
-    return ret;
+    return rb_vm_call0(ec, obj, mid, argc, argv, me, kw_splat);
 }
 
 static VALUE
 vm_caller_setup_arg_block(const rb_execution_context_t *ec, rb_control_frame_t *reg_cfp,
-                          const struct rb_call_info *ci, const rb_iseq_t *blockiseq, const int is_super)
+                          const struct rb_callinfo *ci, const rb_iseq_t *blockiseq, const int is_super)
 {
-    if (ci->flag & VM_CALL_ARGS_BLOCKARG) {
+    if (vm_ci_flag(ci) & VM_CALL_ARGS_BLOCKARG) {
 	VALUE block_code = *(--reg_cfp->sp);
 
 	if (NIL_P(block_code)) {
@@ -1220,7 +897,7 @@ vm_caller_setup_arg_block(const rb_execution_context_t *ec, rb_control_frame_t *
                     rb_ary_push(callback_arg, block_code);
                     rb_ary_push(callback_arg, ref);
                     OBJ_FREEZE_RAW(callback_arg);
-                    func = rb_func_proc_new(refine_sym_proc_call, callback_arg);
+                    func = rb_func_lambda_new(refine_sym_proc_call, callback_arg, 1, UNLIMITED_ARGUMENTS);
 		    rb_hash_aset(ref, block_code, func);
 		}
 		block_code = func;
